
// Single processor.  ad++ OVERALL macros replaced with `LocBlocksIndex'.

#include "ad++.h"
#include <sys/time.h>

const int N = 100 ;
const int NITER = 5000 ;

int wlo [] = {1, 1}, whi [] = {1, 1} ;  // ghost widths for `writeHalo'

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs2 p(1, 1) ;
  ON(p) {
    BlockRange x(N, p.dim(0), LAY_EXTENDED, 1, 1),
               y(N, p.dim(1), LAY_EXTENDED, 1, 1) ;

    float h = 1.0 / (N - 1) ;

    // Initialization (unoptimized):
    //    sol = exact solution
    //    f   = forcing term
    //    u   = approximation solution

    Array2<float> u(x, y), f(x, y), sol(x, y) ;

    const float fac = 4.0 - 2.0 * (cos(h) + cos(3 * h)) ;

    Index i(x), j(y) ;
    OVERALL(i) {
      OVERALL(j) {
	sol(i, j) = sin(3 * (h * i) + h * j) ;
        if(i == 0 || i == N - 1 || j == 0 || j == N - 1)
	  u(i, j) = sol(i, j) ;
        else {
	  u(i, j) = 0.0 ;
	  f(i, j) = fac * sol(i, j) ;
        }
      } ALLOVER(j) ;
    } ALLOVER(i) ;

//    gprintf("initial u:\n") ;
//    aprintf("%7.3f%*N", u, N) ;

    // Main loop

    Index i1(x.triplet(1, N - 2)) ;

    clock() ;

    for(int iter = 0 ; iter < NITER ; iter++) {
      for(int parity = 0 ; parity < 2 ; parity++) {

//      writeHalo(u, wlo, whi) ;

        apgStack.push(&apg) ;
        apg.restrict(i1.dim) ;
        for(i1.beginLocBlk() ; i1.test() ; i1.nextLocBlk()) {
          const int u0_bas = u.str(0).offset(i1) ;
          const int u0_stp = u.str(0).step(i1) ;
          const int u0_len = u.str(0).len() ;

          const int f0_bas = f.str(0).offset(i1) ;
          const int f0_stp = f.str(0).step(i1) ;
          const int f0_len = f.str(0).len() ;

          for(int il = 0 ; il < i1.count ; il++) {
	    Index j1(y.triplet(1 + (1 + i1.glb_bas + i1.glb_stp * il +
                                    parity) % 2, N - 2, 2)) ;
	    apgStack.push(&apg) ;
	    apg.restrict(j1.dim) ;
            for(j1.beginLocBlk() ; j1.test() ; j1.nextLocBlk()) {
	      const int u1_bas = u.str(1).offset(j1) ;
	      const int u1_stp = u.str(1).step(j1) ;
	      const int u1_len = u.str(1).len() ;

	      const int f1_bas = f.str(1).offset(j1) ;
	      const int f1_stp = f.str(1).step(j1) ;
	      const int f1_len = f.str(1).len() ;

	      float* u_dat = u.dat() ;
	      float* f_dat = f.dat() ;

              for(int jl = 0 ; jl < j1.count ; jl++)
		u_dat[u0_bas + u0_stp * il + u1_bas + u1_stp * jl] =
		    0.25 * (f_dat[f0_bas + f0_stp * il +
				  f1_bas + f1_stp * jl] +
			    u_dat[u0_bas + u0_stp * il + u0_len * 1 +
				  u1_bas + u1_stp * jl] +
			    u_dat[u0_bas + u0_stp * il - u0_len * 1 +
				  u1_bas + u1_stp * jl] +
			    u_dat[u0_bas + u0_stp * il +
				  u1_bas + u1_stp * jl + u1_len * 1] +
			    u_dat[u0_bas + u0_stp * il +
				  u1_bas + u1_stp * jl - u1_len * 1]) ;
	    }
            apgStack.pop(&apg) ;
          }
        }
        apgStack.pop(&apg) ;
      }

//      gprintf("iteration = %d\n", iter) ;
//      aprintf("%7.3f%*N", u, N) ;
    }

    clock_t t = clock() ;
  
    printf("elapsed time = %f\n", (float) t / CLOCKS_PER_SEC) ;

//    gprintf("exact solution:\n") ;
//    aprintf("%7.3f%*N", sol, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

